# This script demonstrates how to run HER.
# Install the package
library(devtools)
install_github("jingyuhe/BayesPortfolio")

# Install CVXR package by the following line
# Please use 0.99-2 version. Later versions of CVXR changes function call and does not support the current code
# install_version("CVXR", version="0.99-2")
# This version CVXR might have confict with R 4.9, we recommend R 3.6.3 version

# Other useful packages 
library(BayesPortfolio)
library(lubridate)
library(CVXR)

source("calculate_weight.R")

# either lasso, ridge, rf, pcr, avramov 
method = "HER"

# data type, FF10, FF49, factor20 or sort100
data_type = "factor20"

scaling = TRUE

# prior parameters
prior_precision = 4000
nu_precision = 10000
deltabar_precision = 1000
V_deltabar = 10
nu_sigma = 13
V_sigma = 13

# number of posterior draws
nsamps = 50
# quantile for interval
quant = 0.95
# threshold for the interval
thres = qnorm(1 - (1 - quant) / 2)

# Generate synthetic data
n_assets = 20
n_regressor = 50 # number of regressors
N_Xt_input = n_regressor + 1 # number of columns of the regressor matrix, add one for intercept
# load data
load("demo_data.rda")
# regdata is a list of data, length is number of assets
# in each element regdata[[1]], y is the return, X is a matrix of regressors
# Xt_test is list (lenght is number of assets), each element is matrix of regressors for the testing period
# Here we predict return of 20 stocks for the next 12 months
# Y_test is return, a vector with length number of assets * length of testing period, stacked by assets
# Y_test = (asset1_month1... asset1_month12, ... , asset20_month1, ... , asset20_month12)


# note that the prior we have diag(c(0.00001, rep(V_deltabar, n_regressor))), it means we have almost zero penalty on the intercept
prior_sur = list(betabar = matrix(0, (1 + n_regressor) * n_assets, 1), A = diag(rep(c(0.001, rep(prior_precision, n_regressor)), n_assets)), nu = nu_sigma, V = V_sigma * diag(n_assets), nudeltabeta = nu_precision, Vdeltabeta = diag(c(0.00001, rep(V_deltabar, n_regressor))), deltabar_precision = diag(c(0.00001, rep(deltabar_precision, n_regressor))))

fit = HER(Data = list(regdata = regdata), Mcmc = list(R = nsamps), Prior = prior_sur)

beta = colMeans(fit$betadraw)
N_Xt_input = dim(Xt_test[[1]])[2]

Y_pred = matrix(0, dim(Xt_test[[1]])[1], n_assets)


# calculate covariance of betas
cov_beta = list()
for(mm in 1:n_assets){
    cov_beta[[mm]] = cov(fit$betadraw[, ((mm - 1) * N_Xt_input + 1):(mm * N_Xt_input)])
}

Sigma = matrix(colMeans(fit$Sigmadraw), n_assets, n_assets)

pred_var = matrix(0, dim(Xt_test[[1]])[1], n_assets)

for (tt in 1:dim(Xt_test[[1]])[1]) {
    for (mm in 1:n_assets) {
        next_obs = Xt_test[[mm]][tt,]

        # calculate mean
        Y_pred[tt, mm] = sum(beta[((mm - 1) * N_Xt_input + 1):(mm * N_Xt_input)] * next_obs)
        if (abs(Y_pred[tt, mm]) > 0.2) {
            Y_pred[tt, mm] = sign(Y_pred[tt, mm]) * 0.2
        }
        # calculate predictive interval
        pred_var[tt, mm] = next_obs %*% cov_beta[[mm]] %*% next_obs + Sigma[mm, mm]
    }
}



R2OOS = function(r_real, r_hat, r_bar) {
    denominator_res = (r_real - r_bar) ^ 2
    denominator = sum(denominator_res)
    numerator_res = (r_real - r_hat) ^ 2
    numerator = sum(numerator_res)
    r2oos = 1 - numerator / denominator
    return(r2oos)
}



r2_oos = rep(0, n_assets)

cover = matrix(0, dim(Y_pred)[1], dim(Y_pred)[2])

se_her = matrix(0, 12, n_assets)

se_ave_120 = matrix(0, 12, n_assets)

se_ave_252 = matrix(0, 12, n_assets)

# calculate R2
for (mm in 1:n_assets) {
    # ind = test_port_label == industry_group[mm]
    ind = rep(1:n_assets, each = 12)
    ind = ind == mm
    r2_oos[mm] = R2OOS(Y_test[ind], Y_pred[, mm], Y_bar_120[ind])
    cover[,mm] = (Y_test[ind] >= Y_pred[, mm] - thres * pred_var[, mm]) *  (Y_test[ind] <= Y_pred[, mm] + thres * pred_var[, mm])

    se_her[,mm] = (Y_test[ind] - Y_pred[,mm])^2
    se_ave_120[, mm] = (Y_test[ind] - Y_bar_120[ind])^2
    se_ave_252[, mm] = (Y_test[ind] - Y_bar_252[ind])^2 
}

interval_length = 2 * thres * pred_var

# Calculate allocation weight for assets
Sigma_mean = colMeans(fit$Sigmadraw)
Sigma_mean = matrix(as.vector(Sigma_mean), n_assets, n_assets)

# you may loop over Y_pred[1, ] to Y_pred[12, ] for the next 12 months
calculate_weight_CVXR(Y_pred[1,],Sigma_mean,1.68)
